"""
Contains all functions that access a Request object.
"""

import time

from sqlalchemy import func

from flask_monitoringdashboard.database import Request


def add_request(db_session, duration, endpoint_id, ip):
    """ Adds a request to the database. Returns the id.
    :param db_session: session for the database
    :param duration: duration of the request
    :param endpoint_id: id of the endpoint
    :param ip: IP address of the requester
    :return the id of the request after it was stored in the database
    """
    request = Request(endpoint_id=endpoint_id, duration=duration, ip=ip)
    db_session.add(request)
    db_session.flush()
    return request.id


def get_data_between(db_session, time_from, time_to=None):
    """
        Returns all data in the Request table, for the export data option.
        This function returns all data between time_from and time_to dates. If time_to is not specified, it will
        return all requests after time_from.
    :param db_session: session for the database
    :param time_from: start of the filtering time
    :param time_to: end of the filtering time
    :return list of requests between time_from and time_to
    """
    query = db_session.query(Request).filter(Request.time_requested > time_from)
    if time_to:
        query = query.filter(Request.time_requested <= time_to)
    return query.all()


def get_data(db_session):
    """
    Equivalent function to get_data_from, but returns all data.
    :param db_session: session for the database
    :return list of all requests in the Request table
    """
    return db_session.query(Request).all()


def get_date_of_first_request(db_session):
    """ Returns the date (as unix timestamp) of the first request since FMD was deployed.
    :param db_session: session for the database
    :return time of the first request
    """
    result = db_session.query(Request.time_requested).order_by(Request.time_requested).first()
    if result:
        return int(time.mktime(result[0].timetuple()))
    return -1


def get_date_of_first_request_version(db_session, version):
    """ Returns the date (as unix timestamp) of the first request in the current FMD version.
    :param db_session: session for the database
    :param version: version of the dashboard
    :return time of the first request in that version
    """
    result = db_session.query(Request.time_requested).\
        filter(Request.version_requested == version).\
        order_by(Request.time_requested).first()
    if result:
        return int(time.mktime(result[0].timetuple()))
    return -1


def get_avg_duration(db_session, endpoint_id):
    """ Returns the average duration of all the requests of an endpoint. If there are no requests for that endpoint,
        it returns 0.
    :param db_session: session for the database
    :param endpoint_id: id of the endpoint
    :return average duration
    """
    result = db_session.query(func.avg(Request.duration).label('average')). \
        filter(Request.endpoint_id == endpoint_id).one()
    if result[0]:
        return result[0]
    return 0



import unittest


class TestColors(unittest.TestCase):

    def test_get_color(self):
        from flask_monitoringdashboard import config
        from flask_monitoringdashboard.core.colors import get_color

        config.colors['endpoint'] = '[0, 1, 2]'

        self.assertEqual(get_color('endpoint'), 'rgb(0, 1, 2)')
        self.assertIn(get_color('main'), ['rgb(140, 191, 64)', 'rgb(140.0, 191.0, 64.0)'])



"""
    This file contains all unit tests for the monitor-rules-table in the database. (Corresponding to the file:
    'flask_monitoringdashboard/database/tests.py')
    See info_box.py for how to run the test-cases.
"""

import unittest
import datetime

from flask_monitoringdashboard.database import session_scope, TestEndpoint
from flask_monitoringdashboard.database.tests import get_endpoint_measurements, get_last_tested_times
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, add_fake_test_runs, \
    REQUESTS, NAME, TEST_NAMES, ENDPOINT_ID

NAME2 = 'main2'
SUITE = 3


class TestDBTests(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        add_fake_test_runs()

    def test_add_test_result(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import add_test_result, get_suite_measurements, add_or_update_test
        from flask_monitoringdashboard.database.tested_endpoints import add_endpoint_hit
        from flask_monitoringdashboard import config
        import datetime
        with session_scope() as db_session:
            self.assertEqual(get_suite_measurements(db_session, SUITE), [0])
            for exec_time in REQUESTS:
                for test in TEST_NAMES:
                    add_or_update_test(db_session, test, True, datetime.datetime.utcnow(), config.version)
                    add_test_result(db_session, test, exec_time, datetime.datetime.utcnow(), config.version, SUITE, 0)
                    add_endpoint_hit(db_session, NAME, exec_time, test, config.version, SUITE)
            result = get_suite_measurements(db_session, SUITE)
            self.assertEqual(len(result), len(REQUESTS) * len(TEST_NAMES))

    def test_get_results(self):
        """
            Test whether the function returns the right values.
        """
        self.test_add_test_result()  # can be replaced by test_add_test_result, since this function covers two tests

    def test_get_suites(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import get_test_suites
        self.test_add_test_result()
        with session_scope() as db_session:
            self.assertEqual(2, len(get_test_suites(db_session)))

    def test_get_measurements(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import get_suite_measurements
        with session_scope() as db_session:
            self.assertEqual(get_suite_measurements(db_session, SUITE), [0])
            self.test_add_test_result()
            result = get_suite_measurements(db_session, SUITE)
            self.assertEqual(len(REQUESTS) * 2, len(result))

    def test_get_test_measurements(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.tests import get_endpoint_measurements_job
        with session_scope() as db_session:
            self.assertEqual(1, len(get_endpoint_measurements_job(db_session, NAME, SUITE)))
            self.test_add_test_result()
            result = get_endpoint_measurements_job(db_session, NAME, SUITE)
            self.assertEqual(len(TEST_NAMES) * len(REQUESTS), len(result))

    def test_get_endpoint_measurements(self):
        with session_scope() as db_session:
            self.assertEqual(get_endpoint_measurements(db_session, "1"), [0])
            db_session.add(TestEndpoint(endpoint_id=ENDPOINT_ID, test_id=1, duration=1234, app_version="1.0",
                                        travis_job_id="1", time_added=datetime.datetime.utcnow()))
            db_session.add(TestEndpoint(endpoint_id=ENDPOINT_ID, test_id=1, duration=2345, app_version="1.0",
                                        travis_job_id="1", time_added=datetime.datetime.utcnow()))
            self.assertEqual(get_endpoint_measurements(db_session, "1"), [1234, 2345])

    def test_get_last_tested_times(self):
        with session_scope() as db_session:
            self.assertEqual(get_last_tested_times(db_session), [])
            db_session.add(TestEndpoint(endpoint_id=ENDPOINT_ID, test_id=1, duration=1234, app_version="1.0",
                                        travis_job_id="1", time_added=datetime.datetime.utcnow()))
            self.assertNotEqual(get_last_tested_times(db_session), [])



"""
    The app tracks the performance of various endpoints over time.
    To bind, use the following lines of code:
        import dashboard
        from flask import Flask
        ...
        app = Flask(__name__)
        ...
        dashboard.bind(app)

    The dashboard with the results that are collected can be found at:
        localhost:5000/dashboard
"""

import os

from flask import Blueprint

from flask_monitoringdashboard.core.config import Config

config = Config()
user_app = None


# get current location of the project
def loc():
    return os.path.abspath(os.path.dirname(__file__)) + '/'


# define the blueprint
blueprint = Blueprint('dashboard', __name__, template_folder=loc() + 'templates')


def bind(app):
    """
        Binding the app to this object should happen before importing the routing-
        methods below. Thus, the importing statement is part of this function.
        :param app: the app for which the performance has to be tracked
    """
    from flask_monitoringdashboard.core.logger import log
    assert app is not None
    global user_app, blueprint
    user_app = app

    # Provide a secret-key for using WTF-forms
    if not user_app.secret_key:
        log('WARNING: You should provide a security key.')
        user_app.secret_key = 'my-secret-key'

    import os
    # Only initialize unit test logging when running on Travis.
    if 'TRAVIS' in os.environ:
        import datetime
        from flask import request

        @user_app.before_first_request
        def log_current_version():
            """
            Logs the version of the user app that is currently being tested.
            :return:
            """
            home = os.path.expanduser("~")
            with open(home + '/app_version.log', 'w') as app_log:
                app_log.write(config.version)

        @user_app.before_request
        def log_start_endpoint_hit():
            """
            Add log_start_endpoint_hit as before_request function that logs the start of the endpoint hit.
            :return:
            """
            hit_time_stamp = str(datetime.datetime.utcnow())
            home = os.path.expanduser("~")
            with open(home + '/start_endpoint_hits.log', 'a') as hits_log:
                hits_log.write('"{}","{}"\n'.format(hit_time_stamp, request.endpoint))

        @user_app.after_request
        def log_finish_endpoint_hit(response):
            """
            Add log_finish_endpoint_hit as after_request function that logs the finish of the endpoint hit.
            :param response: the response object that the actual endpoint returns
            :return: the unchanged response of the original endpoint
            """
            hit_time_stamp = str(datetime.datetime.utcnow())
            home = os.path.expanduser("~")
            with open(home + '/finish_endpoint_hits.log', 'a') as hits2_log:
                hits2_log.write('"{}","{}"\n'.format(hit_time_stamp, request.endpoint))

            return response

    # Add all route-functions to the blueprint
    import flask_monitoringdashboard.views

    # Add wrappers to the endpoints that have to be monitored
    from flask_monitoringdashboard.core.measurement import init_measurement
    blueprint.before_app_first_request(init_measurement)

    # register the blueprint to the app
    app.register_blueprint(blueprint, url_prefix='/' + config.link)



from flask_monitoringdashboard import blueprint, config
from flask import redirect, request, session, render_template, url_for
from flask_monitoringdashboard.core.forms import Login
from flask_monitoringdashboard.core.auth import check_login, on_logout


MAIN_PAGE = 'dashboard.index'


@blueprint.route('/login', methods=['GET', 'POST'])
def login():
    """
    User for logging into the system. The POST-request checks whether the logging is valid.
    If this is the case, the user is redirected to the main page.
    :return:
    """
    if session.get(config.link + '_logged_in'):
        return redirect(url_for(MAIN_PAGE))

    form = Login()
    if request.method == 'POST' and form.validate():
        if not check_login(name=form.name.data, password=form.password.data):
            form.name.errors.append('Incorrect username or password')
        else:
            return redirect(url_for(MAIN_PAGE))
    return render_template('fmd_login.html', form=form)


@blueprint.route('/logout')
def logout():
    """
    Remove the session variables from the user.
    Redirect the user to the login page.
    :return:
    """
    return on_logout()



import datetime

import numpy
import plotly.graph_objs as go
from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.forms import get_daterange_form
from flask_monitoringdashboard.core.plot import get_layout, get_figure, heatmap as plot_heatmap
from flask_monitoringdashboard.core.info_box import get_plot_info
from flask_monitoringdashboard.core.timezone import to_utc_datetime, to_local_datetime
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.endpoint import get_num_requests

TITLE = 'Hourly API Utilization'

AXES_INFO = '''The X-axis presents a number of days. The Y-axis presents every hour of 
the day.'''

CONTENT_INFO = '''The color of the cell presents the number of requests that the application received 
in a single hour. The darker the cell, the more requests it has processed. This information can be used 
to validate on which moment of the day the Flask application processes to most requests.'''


@blueprint.route('/hourly_load', methods=['GET', 'POST'])
@secure
def hourly_load():
    form = get_daterange_form()
    return render_template('fmd_dashboard/graph.html', form=form, graph=hourly_load_graph(form), title=TITLE,
                           information=get_plot_info(AXES_INFO, CONTENT_INFO))


def hourly_load_graph(form, endpoint_id=None):
    """
    Return HTML string for generating a Heatmap.
    :param form: A SelectDateRangeForm, which is used to filter the selection
    :param endpoint_id: optionally, filter the data on a specific endpoint
    :return: HTML code with the graph
    """
    # list of hours: 0:00 - 23:00
    hours = ['0{}:00'.format(h) for h in range(0, 10)] + ['{}:00'.format(h) for h in range(10, 24)]
    days = form.get_days()

    # create empty 2D-list: [hour][day]
    heatmap_data = numpy.zeros((len(hours), len(days)))

    # add data from database to heatmap_data
    start_datetime = to_utc_datetime(datetime.datetime.combine(form.start_date.data, datetime.time(0, 0, 0, 0)))
    end_datetime = to_utc_datetime(datetime.datetime.combine(form.end_date.data, datetime.time(23, 59, 59)))

    with session_scope() as db_session:
        for time, count in get_num_requests(db_session, endpoint_id, start_datetime, end_datetime):
            parsed_time = datetime.datetime.strptime(time, '%Y-%m-%d %H:%M:%S')
            day_index = (parsed_time - start_datetime).days
            hour_index = int(to_local_datetime(parsed_time).strftime('%H'))
            heatmap_data[hour_index][day_index] = count

    start_datetime = to_local_datetime(start_datetime - datetime.timedelta(days=1)).strftime('%Y-%m-%d 12:00:00')
    end_datetime = to_local_datetime(form.end_date.data).strftime('%Y-%m-%d 12:00:00')

    layout = get_layout(
        xaxis=go.XAxis(range=[start_datetime, end_datetime])
    )
    return get_figure(layout, [plot_heatmap(x=days, y=hours, z=heatmap_data)])



"""
    Contains all plots that are visible in the Dashboard
"""
import math

import plotly.graph_objs as go

from flask_monitoringdashboard.core.colors import get_color
from flask_monitoringdashboard.core.plot.util import add_default_value

BUBBLE_SIZE_RATIO = 2500


def heatmap(x, y, z, **kwargs):
    """
    :param x: list for labels of the x-values
    :param y: list for lables of the y-values
    :param z: 2D list with the actual data, encoded as: z[x-index][y-index]
    :param kwargs: additional arguments for the heatmap
    :return: a Heatmap that can be used for generating a Plotly figure :func:`get_figure`
    """
    kwargs = add_default_value('colorscale',
                               [[0, 'rgb(255, 255, 255)'], [0.01, 'rgb(240,240,240)'], [1, 'rgb(1, 1, 1)']], **kwargs)
    return go.Heatmap(x=x, y=y, z=z, **kwargs)


def boxplot(values, **kwargs):
    """
    :param values: values for the boxplot
    :param kwargs: additional arguments for the boxplot
    :return: A boxplot that can be used for generating a Plotly figure :func:`get_figure`
    """
    if 'name' in kwargs.keys():
        kwargs = add_default_value('marker', {'color': get_color(kwargs.get('name', ''))}, **kwargs)
    if 'label' in kwargs.keys():
        kwargs = add_default_value('name', kwargs.get('label', ''))
    kwargs = add_default_value('x', value=values, **kwargs)
    return go.Box(**kwargs)


def barplot(x, y, name, **kwargs):
    """
    :param x:
    :param y:
    :param name:
    :param kwargs: additional arguments
    :return: A barplot that can be used for generating a Plotly figure :func:`get_figure`
    """
    return go.Bar(
        x=x,
        y=y,
        name=name,
        orientation='h',
        marker={'color': get_color(name)},
        **kwargs
    )


def scatter(**kwargs):
    """

    :param kwargs: additional arguments for the scatterplot
    :return: a scatterplot that can be used for generating a Plotly figure :func:`get_figure`
    """
    return go.Scatter(**kwargs)


def get_average_bubble_size(data):
    """
    :param data: a list with lists: [[a, b, c], [d, e, f]]
    :return: a constant for the bubble size
    """
    try:
        return math.sqrt(max([max([r for r in row]) for row in data])) / BUBBLE_SIZE_RATIO
    except ValueError:
        return BUBBLE_SIZE_RATIO



from flask_monitoringdashboard.database import Endpoint, Test, TestEndpoint


def add_endpoint_hit(db_session, endpoint, time, test, version, job_id):
    """
    Adds an endpoint hit to the database.
    :param db_session: Session to connect to the db.
    :param endpoint: Name of the endpoint that was hit.
    :param time: Execution time in ms of the endpoint hit.
    :param test: Name of the test that caused the hit.
    :param version: Version of the user app in which the hit occurred.
    :param job_id: Travis job ID in which the hit occurred.
    """
    endpoint_id = db_session.query(Endpoint.id).filter(Endpoint.name == endpoint).first().id
    test_id = db_session.query(Test.id).filter(Test.name == test).first().id
    db_session.add(TestEndpoint(endpoint_id=endpoint_id, test_id=test_id, duration=time, app_version=version,
                                travis_job_id=job_id))


def get_tested_endpoint_names(db_session):
    """
    Returns the names of all of the endpoint for which test data is collected.
    :param db_session:
    :return list of strings
    """
    results = db_session.query(Endpoint.name).join(TestEndpoint.endpoint).\
        group_by(Endpoint.name).all()
    return [result[0] for result in results]



from sqlalchemy import func, distinct

from flask_monitoringdashboard.database import Request, StackLine, TestResult, TestEndpoint


def count_rows(db_session, column, *criterion):
    """
    Count the number of rows of a specified column.
    :param db_session: session for the database
    :param column: column to count
    :param criterion: where-clause of the query
    :return: number of rows
    """
    result = db_session.query(func.count(distinct(column))).filter(*criterion).first()
    if result:
        return result[0]
    return 0


def count_users(db_session, endpoint_id):
    """
    :param db_session: session for the database
    :param endpoint_id: id of the endpoint
    :return: The number of distinct users that have requested this endpoint
    """
    return count_rows(db_session, Request.group_by, Request.endpoint_id == endpoint_id)


def count_ip(db_session, endpoint_id):
    """
    :param db_session: session for the database
    :param endpoint_id: id of the endpoint
    :return: The number of distinct users that have requested this endpoint
    """
    return count_rows(db_session, Request.ip, Request.endpoint_id == endpoint_id)


def count_versions(db_session):
    """
    :param db_session: session for the database
    :return: The number of distinct versions that are used
    """
    return count_rows(db_session, Request.version_requested)


def count_test_builds(db_session):
    """
    :param db_session: session for the database
    :return: The number of Travis builds that are available
    """
    return count_rows(db_session, TestResult.travis_job_id)


def count_builds_endpoint(db_session):
    """
    :param db_session: session for the database
    :return: The number of Travis builds that are available
    """
    return count_rows(db_session, TestEndpoint.travis_job_id)


def count_versions_endpoint(db_session, endpoint_id):
    """
    :param db_session: session for the database
    :param endpoint_id: id of the endpoint
    :return: The number of distinct versions that are used for this endpoint
    """
    return count_rows(db_session, Request.version_requested, Request.endpoint_id == endpoint_id)


def count_requests(db_session, endpoint_id, *where):
    """
    Return the number of hits for a specific endpoint (possible with more filter arguments).
    :param db_session: session for the database
    :param endpoint_id: id of the endpoint
    :param where: additional arguments
    """
    return count_rows(db_session, Request.id, Request.endpoint_id == endpoint_id, *where)


def count_total_requests(db_session, *where):
    """
    Return the number of total hits
    :param db_session: session for the database
    :param where: additional arguments
    """
    return count_rows(db_session, Request.id, *where)


def count_outliers(db_session, endpoint_id):
    """
    :param db_session: session for the database
    :param endpoint_id: id of the endpoint
    :return: An integer with the number of rows in the Outlier-table.
    """
    return count_rows(db_session, Request.id, Request.endpoint_id == endpoint_id, Request.outlier)


def count_profiled_requests(db_session, endpoint_id):
    """
    Count the number of profiled requests for a certain endpoint
    :param db_session: session for the database
    :param endpoint_id: id of the endpoint
    :return: An integer
    """
    count = db_session.query(func.count(distinct(StackLine.request_id))). \
        filter(Request.endpoint_id == endpoint_id).\
        join(Request.stack_lines).first()
    if count:
        return count[0]
    return 0



"""
    Creates the database. 
    For information about how to access the database via a session-variable, see: session_scope() 
"""
import datetime
import random
import time
from contextlib import contextmanager

from sqlalchemy import Column, Integer, String, DateTime, create_engine, Float, Boolean, TEXT, ForeignKey, exc
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship, scoped_session

from flask_monitoringdashboard import config
from flask_monitoringdashboard.core.group_by import get_group_by

Base = declarative_base()


class Endpoint(Base):
    """ Table for storing information about the endpoints. """
    __tablename__ = '{}Endpoint'.format(config.table_prefix)
    id = Column(Integer, primary_key=True)
    # name of the endpoint
    name = Column(String(250), unique=True, nullable=False)
    # 0 - disabled, 1 - performance, 2 - profiler, 3 - profiler + outliers
    monitor_level = Column(Integer, default=config.monitor_level)
    # the time when the endpoint was added
    time_added = Column(DateTime, default=datetime.datetime.utcnow)
    # the version when the endpoint was added
    version_added = Column(String(100), default=config.version)
    # the time the endpoint was last used
    last_requested = Column(DateTime)


class Request(Base):
    """ Table for storing measurements of requests. """
    __tablename__ = '{}Request'.format(config.table_prefix)
    id = Column(Integer, primary_key=True)
    endpoint_id = Column(Integer, ForeignKey(Endpoint.id))
    # the processing time of the request
    duration = Column(Float, nullable=False)
    # the time when the request was handled
    time_requested = Column(DateTime, default=datetime.datetime.utcnow)
    # the version when the request was handled
    version_requested = Column(String(100), default=config.version)
    # a criteria by which the requests can be grouped
    group_by = Column(String(100), default=get_group_by)
    # the ip address of the requester
    ip = Column(String(100), nullable=False)

    endpoint = relationship(Endpoint)
    stack_lines = relationship('StackLine', back_populates='request')
    outlier = relationship('Outlier', uselist=False, back_populates='request')


class Outlier(Base):
    """ Table for storing information about outliers. """
    __tablename__ = '{}Outlier'.format(config.table_prefix)
    id = Column(Integer, primary_key=True)
    request_id = Column(Integer, ForeignKey(Request.id))
    # http headers of the request
    request_header = Column(TEXT)
    # http environment of the request
    request_environment = Column(TEXT)
    # url of the request
    request_url = Column(String(2100))
    # cpu utilization of the server when handling the request
    cpu_percent = Column(String(150))
    # memory utilization of the server when handling the request
    memory = Column(TEXT)
    # the stacktrace of the request
    stacktrace = Column(TEXT)

    request = relationship(Request, back_populates='outlier')


class CodeLine(Base):
    __tablename__ = '{}CodeLine'.format(config.table_prefix)
    """ Table for storing the text of a StackLine. """
    id = Column(Integer, primary_key=True)
    # quadruple (filename, line_number, function_name, code) that uniquely identifies a statement in the code
    filename = Column(String(250), nullable=False)
    line_number = Column(Integer, nullable=False)
    function_name = Column(String(250), nullable=False)
    code = Column(String(250), nullable=False)


class StackLine(Base):
    """ Table for storing lines of execution paths of calls. """
    __tablename__ = '{}StackLine'.format(config.table_prefix)
    request_id = Column(Integer, ForeignKey(Request.id), primary_key=True)
    code_id = Column(Integer, ForeignKey(CodeLine.id))
    # position in the flattened stack tree
    position = Column(Integer, primary_key=True)
    # level in the tree
    indent = Column(Integer, nullable=False)
    # time spent on that line
    duration = Column(Float, nullable=False)

    request = relationship(Request, back_populates='stack_lines')
    code = relationship(CodeLine)


class Test(Base):
    """ Stores all of the tests that exist in the project. """
    __tablename__ = '{}Test'.format(config.table_prefix)
    id = Column(Integer, primary_key=True)
    # name of the test
    name = Column(String(250), unique=True)
    # true - the test passed, false - otherwise
    passing = Column(Boolean, nullable=False)
    # time of the most recent run of the test
    last_tested = Column(DateTime, default=datetime.datetime.utcnow)
    # version when the test was added
    version_added = Column(String(100), nullable=False)
    # time when the test was added
    time_added = Column(DateTime, default=datetime.datetime.utcnow)


class TestResult(Base):
    """ Stores unit test performance results obtained from Travis. """
    __tablename__ = '{}TestResult'.format(config.table_prefix)
    id = Column(Integer, primary_key=True)
    test_id = Column(Integer, ForeignKey(Test.id))
    # time it took to run the test
    duration = Column(Float, nullable=False)
    # time when the test was run
    time_added = Column(DateTime, default=datetime.datetime.utcnow)
    # version when the test was run
    app_version = Column(String(100), nullable=False)
    # id of the travis job running the test
    travis_job_id = Column(String(10), nullable=False)
    # number of the travis run
    run_nr = Column(Integer, nullable=False)

    test = relationship(Test)


class TestEndpoint(Base):
    """ Stores the endpoint hits that came from unit tests. """
    __tablename__ = '{}TestEndpoint'.format(config.table_prefix)
    id = Column(Integer, primary_key=True)
    endpoint_id = Column(Integer, ForeignKey(Endpoint.id))
    test_id = Column(Integer, ForeignKey(Test.id))
    # duration of the tests for an endpoint
    duration = Column(Float, nullable=False)
    # version when the endpoint was tested
    app_version = Column(String(100), nullable=False)
    # id of the travis job running the test
    travis_job_id = Column(String(10), nullable=False)
    # time when the endpoint was tested
    time_added = Column(DateTime, default=datetime.datetime.utcnow)

    endpoint = relationship(Endpoint)
    test = relationship(Test)


# define the database
engine = create_engine(config.database_name)

# creates all tables in the database
Base.metadata.create_all(engine)
Base.metadata.bind = engine
DBSession = sessionmaker(bind=engine)


@contextmanager
def session_scope():
    """
    When accessing the database, use the following syntax:
        with session_scope() as db_session:
            db_session.query(...)
            
    :return: the session for accessing the database
    """
    session_obj = scoped_session(DBSession)
    session = session_obj()
    try:
        yield session
        session.commit()
    except exc.OperationalError:
        session.rollback()
        time.sleep(0.5 + random.random())
        session.commit()
    except Exception as e:
        session.rollback()
        print('No commit has been made, due to the following error: {}'.format(e))
    finally:
        session.close()


def get_tables():
    return [Endpoint, Request, Outlier, StackLine, CodeLine, Test, TestResult, TestEndpoint]



"""
    This file contains all unit tests that count a number of results in the database. (Corresponding to the file:
    'flask_monitoringdashboard/database/count.py')
    See info_box.py for how to run the test-cases.
"""

import unittest

from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.code_line import get_code_line

FN = 'filename'
LN = 42
FUN = 'fun'
CODE = 'code'


class TestCodeLine(unittest.TestCase):

    def test_get_code_line(self):
        with session_scope() as db_session:
            code_line1 = get_code_line(db_session, FN, LN, FUN, CODE)
            code_line2 = get_code_line(db_session, FN, LN, FUN, CODE)
            self.assertEqual(code_line1.id, code_line2.id)
            self.assertEqual(code_line1.function_name, code_line2.function_name)
            self.assertEqual(code_line1.filename, code_line2.filename)
            self.assertEqual(code_line1.line_number, code_line2.line_number)
            self.assertEqual(code_line1.code, code_line2.code)



from flask_monitoringdashboard.core.profiler.baseProfiler import BaseProfiler
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.endpoint import update_last_accessed
from flask_monitoringdashboard.database.request import add_request


class PerformanceProfiler(BaseProfiler):
    """
    Used for updating the performance and utilization of the endpoint in the database.
    Used when monitoring-level == 1
    """

    def __init__(self, endpoint, ip, duration):
        super(PerformanceProfiler, self).__init__(endpoint)
        self._ip = ip
        self._duration = duration * 1000  # Conversion from sec to ms
        self._endpoint = endpoint

    def run(self):
        with session_scope() as db_session:
            update_last_accessed(db_session, endpoint_name=self._endpoint.name)
            add_request(db_session, duration=self._duration, endpoint_id=self._endpoint.id, ip=self._ip)



import datetime

from flask import request
from flask_wtf import FlaskForm
from wtforms import validators, SubmitField
from wtforms.fields.html5 import DateField

from flask_monitoringdashboard.core.timezone import to_local_datetime

DATE_FORMAT = '%Y-%m-%d'


class SelectDateRangeForm(FlaskForm):
    """ Used for selecting two dates, which together specify a range. """
    start_date = DateField('Start date', format=DATE_FORMAT, validators=[validators.data_required()])
    end_date = DateField('End date', format=DATE_FORMAT, validators=[validators.data_required()])
    submit = SubmitField('Update')
    title = 'Select the time interval'

    def get_days(self):
        """
        :return: A list with datetime.date object from form.start_date to (including) form.end_date
        """
        delta = self.end_date.data - self.start_date.data
        return [self.start_date.data + datetime.timedelta(days=i) for i in range(delta.days + 1)]

    def content(self):
        return '''
          <div class="row">
            <div class="col-sm-4"><i class="fa fa-calendar"></i> {} </div>
            <div class="col-sm-4"><i class="fa fa-calendar"></i> {} </div>
          </div>
          <div class="row">
            <div class="col-sm-4"> {} </div>
            <div class="col-sm-4"> {} </div>
            <div class="col-sm-4"> {} </div>
          </div>'''.format(self.start_date.label, self.end_date.label,
                           self.start_date(class_="form-control", required=True),
                           self.end_date(class_="form-control", required=True),
                           self.submit(class_="btn btn-primary btn-block"))


def get_daterange_form(num_days=20):
    """
    Returns a SelectDateRangeForm with two dates:
    - end_date is today
    - start_date is the today - numdays
    :param num_days: the date for the start_date
    :return: A SelectDateRangeForm object with the required logic
    """
    form = SelectDateRangeForm(request.form)
    if form.validate():
        if form.start_date.data > form.end_date.data:
            form.start_date.data, form.end_date.data = form.end_date.data, form.start_date.data
    else:
        form.end_date.data = to_local_datetime(datetime.datetime.utcnow()).date()
        form.start_date.data = form.end_date.data - datetime.timedelta(days=num_days)
    return form



from sqlalchemy.orm.exc import NoResultFound

from flask_monitoringdashboard.database import CodeLine


def get_code_line(db_session, fn, ln, name, code):
    """
    Get a CodeLine object from a given quadruple of fn, ln, name, code. If the CodeLine object doesn't already exist,
    a new one is created in the database.
    :param db_session: session for the database
    :param fn: filename (string)
    :param ln: line_number of the code (int)
    :param name: function name (string)
    :param code: line of code (string)
    :return: a CodeLine object
    """
    try:
        result = db_session.query(CodeLine). \
            filter(CodeLine.filename == fn, CodeLine.line_number == ln, CodeLine.function_name == name,
                   CodeLine.code == code).one()
    except NoResultFound:
        result = CodeLine(filename=fn, line_number=ln, function_name=name, code=code)
        db_session.add(result)
        db_session.flush()
    return result



import unittest

from flask_monitoringdashboard.database.stack_line import add_stack_line, get_profiled_requests, \
    get_grouped_profiled_requests

from flask_monitoringdashboard.database import session_scope, StackLine
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, ENDPOINT_ID


class TestStackLine(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    def test_add_stackline(self):
        with session_scope() as db_session:
            self.assertEqual(db_session.query(StackLine).all(), [])
            add_stack_line(db_session, request_id=1, position=0, indent=1, duration=1, code_line="code")
            self.assertNotEqual(db_session.query(StackLine).all(), [])

    def test_get_profiled_requests(self):
        with session_scope() as db_session:
            self.assertEqual(get_profiled_requests(db_session, endpoint_id=ENDPOINT_ID, offset=0, per_page=10), [])
            add_stack_line(db_session, request_id=ENDPOINT_ID, position=0, indent=1, duration=1, code_line="code")
            self.assertNotEqual(get_profiled_requests(db_session, endpoint_id=ENDPOINT_ID, offset=0, per_page=10), [])

    def test_get_grouped_profiled_requests(self):
        with session_scope() as db_session:
            self.assertEqual(get_grouped_profiled_requests(db_session, endpoint_id=ENDPOINT_ID), [])
            add_stack_line(db_session, request_id=ENDPOINT_ID, position=0, indent=1, duration=1, code_line="code")
            self.assertNotEqual(get_grouped_profiled_requests(db_session, endpoint_id=ENDPOINT_ID), [])



from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.forms import get_daterange_form
from flask_monitoringdashboard.core.plot import barplot, get_figure, get_layout
from flask_monitoringdashboard.core.info_box import get_plot_info
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.count_group import count_requests_per_day, get_value
from flask_monitoringdashboard.database.endpoint import get_endpoints

TITLE = 'Daily API Utilization'

AXES_INFO = '''The X-axis presents the amount of requests. The Y-axis presents a number 
of days'''

CONTENT_INFO = '''This graph presents a horizontal stacked barplot. Each endpoint is represented by a 
color. In the legend on the right, you can disable a certain endpoint by clicking on it. You can also
show in the information of a single endpoint by double clicking that endpoint in the legend. The 
information from this graph can be used to see on which days (a subset of) the endpoints are used to 
most.'''


@blueprint.route('/requests', methods=['GET', 'POST'])
@secure
def requests():
    form = get_daterange_form(num_days=10)
    return render_template('fmd_dashboard/graph.html', form=form, graph=requests_graph(form), title=TITLE,
                           information=get_plot_info(AXES_INFO, CONTENT_INFO))


def requests_graph(form):
    """
    Returns a horizontal box plot with the number of requests per day.
    :param form: must be the form that is obtained by get_daterange_form
    :return:
    """
    days = form.get_days()
    with session_scope() as db_session:
        hits = count_requests_per_day(db_session, days)
        data = [barplot(x=[get_value(hits_day, end.id) for hits_day in hits], y=days, name=end.name)
                for end in get_endpoints(db_session)]
    layout = get_layout(
        barmode='stack',
        height=350 + 40 * len(days),
        showlegend=True,
        xaxis={'title': 'Number of requests'},
        yaxis={'type': 'category'}
    )

    return get_figure(layout=layout, data=data)



from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.forms import get_daterange_form
from flask_monitoringdashboard.core.info_box import get_plot_info
from flask_monitoringdashboard.core.utils import get_endpoint_details
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.views.dashboard.heatmap import hourly_load_graph, TITLE, AXES_INFO

CONTENT_INFO = '''The color of the cell presents the number of requests that the application received 
in a single hour. The darker the cell, the more requests it has processed. This information can be used 
to validate on which moment of the day this endpoint processes to most requests.'''


@blueprint.route('/endpoint/<int:endpoint_id>/hourly_load', methods=['GET', 'POST'])
@secure
def endpoint_hourly_load(endpoint_id):
    form = get_daterange_form()
    with session_scope() as db_session:
        details = get_endpoint_details(db_session, endpoint_id)
    title = '{} for {}'.format(TITLE, details['endpoint'])
    graph = hourly_load_graph(form, endpoint_id)

    return render_template('fmd_dashboard/graph-details.html', form=form, details=details, graph=graph, title=title,
                           information=get_plot_info(AXES_INFO, CONTENT_INFO))



from sqlalchemy import desc
from sqlalchemy.orm import joinedload

from flask_monitoringdashboard.database import Outlier, Request


def add_outlier(db_session, request_id, cpu_percent, memory, stacktrace, request):
    """
    Adds an Outlier object in the database.
    :param db_session: session for the database
    :param request_id: id of the request
    :param cpu_percent: cpu load of the server when processing the request
    :param memory: memory load of the server when processing the request
    :param stacktrace: stack trace of the request
    :param request: triple containing the headers, environment and url
    """
    headers, environ, url = request
    outlier = Outlier(request_id=request_id, request_header=headers,
                      request_environment=environ,
                      request_url=url, cpu_percent=cpu_percent,
                      memory=memory, stacktrace=stacktrace)
    db_session.add(outlier)


def get_outliers_sorted(db_session, endpoint_id, offset, per_page):
    """
    Gets a list of Outlier objects for a certain endpoint, sorted by most recent request time
    :param db_session: session for the database
    :param endpoint_id: id of the endpoint for filtering the requests
    :param offset: number of items to skip
    :param per_page: number of items to return
    :return list of Outlier objects of a specific endpoint
    """
    result = db_session.query(Outlier).\
        join(Outlier.request). \
        options(joinedload(Outlier.request).joinedload(Request.endpoint)). \
        filter(Request.endpoint_id == endpoint_id). \
        order_by(desc(Request.time_requested)). \
        offset(offset).limit(per_page).all()
    db_session.expunge_all()
    return result


def get_outliers_cpus(db_session, endpoint_id):
    """
    Gets list of CPU loads of all outliers of a certain endpoint
    :param db_session: session for the database
    :param endpoint_id: id of the endpoint
    :return list of cpu percentages as strings
    """
    outliers = db_session.query(Outlier.cpu_percent).\
        join(Outlier.request).\
        filter(Request.endpoint_id == endpoint_id).all()
    return [outlier[0] for outlier in outliers]


def get_all_outliers(db_session):
    """
    :param db_session: session for the database
    :return list of all outliers in the Outlier table
    """
    return db_session.query(Outlier).all()



import datetime

from flask import make_response

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import admin_secure
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.outlier import get_all_outliers
from flask_monitoringdashboard.database.request import get_data

REQUESTS_COLUMNS = ['id', 'endpoint_id', 'duration', 'time_requested',
                    'version_requested', 'group_by', 'ip']
OUTLIER_COLUMNS = ['id', 'request_id', 'request_header', 'request_environment',
                   'request_url', 'cpu_percent', 'memory', 'stacktrace']


@blueprint.route('/download-requests')
@admin_secure
def download_requests():
    csv = ','.join(REQUESTS_COLUMNS) + '\n'
    with session_scope() as db_session:
        for entry in get_data(db_session):
            csv += ','.join([str(entry.__getattribute__(c)) for c in REQUESTS_COLUMNS]) + '\n'

    response = make_response(csv)
    response.headers["Content-Disposition"] = "attachment; filename=requests_{0}.csv".format(
        str(datetime.datetime.utcnow()).replace(" ", "_").replace(":", "-")[:19])
    return response


@blueprint.route('/download-outliers')
@admin_secure
def download_outliers():
    csv = ','.join(OUTLIER_COLUMNS) + '\n'
    with session_scope() as db_session:
        for entry in get_all_outliers(db_session):
            data = ','.join([str(entry.__getattribute__(c)) for c in OUTLIER_COLUMNS])
            data = ' '.join(data.split())  # remove newlines
            csv += data + '\n'

    response = make_response(csv)
    response.headers["Content-Disposition"] = "attachment; filename=outliers_{0}.csv".format(
        str(datetime.datetime.utcnow()).replace(" ", "_").replace(":", "-")[:19])
    return response



import datetime

from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure, is_admin
from flask_monitoringdashboard.core.colors import get_color
from flask_monitoringdashboard.core.timezone import to_local_datetime, to_utc_datetime
from flask_monitoringdashboard.core.utils import get_details
from flask_monitoringdashboard.database import Request, session_scope
from flask_monitoringdashboard.database.count_group import count_requests_group, get_value
from flask_monitoringdashboard.database.data_grouped import get_endpoint_data_grouped
from flask_monitoringdashboard.database.endpoint import get_last_requested, get_endpoints


@blueprint.route('/overview')
@secure
def overview():
    week_ago = datetime.datetime.utcnow() - datetime.timedelta(days=7)
    now_local = to_local_datetime(datetime.datetime.utcnow())
    today_local = now_local.replace(hour=0, minute=0, second=0, microsecond=0)
    today_utc = to_utc_datetime(today_local)

    result = []
    with session_scope() as db_session:
        from numpy import median

        hits_today = count_requests_group(db_session, Request.time_requested > today_utc)
        hits_week = count_requests_group(db_session, Request.time_requested > week_ago)
        hits = count_requests_group(db_session)

        median_today = get_endpoint_data_grouped(db_session, median, Request.time_requested > today_utc)
        median_week = get_endpoint_data_grouped(db_session, median, Request.time_requested > week_ago)
        median = get_endpoint_data_grouped(db_session, median)
        access_times = get_last_requested(db_session)

        for endpoint in get_endpoints(db_session):
            result.append({
                'id': endpoint.id,
                'name': endpoint.name,
                'color': get_color(endpoint.name),
                'hits-today': get_value(hits_today, endpoint.id),
                'hits-week': get_value(hits_week, endpoint.id),
                'hits-overall': get_value(hits, endpoint.id),
                'median-today': get_value(median_today, endpoint.id),
                'median-week': get_value(median_week, endpoint.id),
                'median-overall': get_value(median, endpoint.id),
                'last-accessed': get_value(access_times, endpoint.name, default=None)
            })
        version = get_details(db_session)['dashboard-version']
    return render_template('fmd_dashboard/overview.html', result=result, is_admin=is_admin(),
                           title='Dashboard Overview', version=version)



import unittest


class TestGroupBy(unittest.TestCase):

    def test_get_group_by(self):
        """
            Test whether the group_by returns the right result
        """
        from flask_monitoringdashboard import config
        from flask_monitoringdashboard.core.group_by import get_group_by
        config.group_by = lambda: 3
        self.assertEqual(get_group_by(), '3')

        config.group_by = (lambda: 'User', lambda: 3.0)
        self.assertEqual(get_group_by(), '(User,3.0)')

        config.group_by = lambda: lambda: '1234'
        self.assertEqual(get_group_by(), '1234')

    def test_recursive_group_by(self):
        from flask_monitoringdashboard.core.group_by import recursive_group_by

        class Object(object):
            def __str__(self):
                return 'object'

        self.assertEqual(recursive_group_by(Object()), 'object')


from flask import render_template
from flask_paginate import get_page_args, Pagination

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.timezone import to_local_datetime
from flask_monitoringdashboard.core.utils import get_endpoint_details
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.count import count_profiled_requests
from flask_monitoringdashboard.database.stack_line import get_profiled_requests


def get_body(index, stack_lines):
    """
    Return the lines (as a list) that belong to the line given in the index
    :param index: integer, between 0 and length(lines)
    :param stack_lines: all lines belonging to a certain request. Every element in this list is an StackLine-obj.
    :return: an empty list if the index doesn't belong to a function. If the list is not empty, it denotes the body of
    the given line (by the index).
    """
    body = []
    indent = stack_lines[index].indent
    index += 1
    while index < len(stack_lines) and stack_lines[index].indent > indent:
        body.append(index)
        index += 1
    return body


@blueprint.route('/endpoint/<endpoint_id>/profiler')
@secure
def profiler(endpoint_id):
    page, per_page, offset = get_page_args(page_parameter='page', per_page_parameter='per_page')
    with session_scope() as db_session:
        details = get_endpoint_details(db_session, endpoint_id)
        requests = get_profiled_requests(db_session, endpoint_id, offset, per_page)

        total = count_profiled_requests(db_session, endpoint_id)
    pagination = Pagination(page=page, per_page=per_page, total=total, format_number=True,
                            css_framework='bootstrap4', format_total=True, record_name='profiled requests')

    body = {}  # dict with the request.id as a key, and the values is a list for every stack_line.
    for request in requests:
        request.time_requested = to_local_datetime(request.time_requested)
        body[request.id] = [get_body(index, request.stack_lines) for index, _ in enumerate(request.stack_lines)]

    return render_template('fmd_dashboard/profiler.html', details=details, requests=requests, pagination=pagination,
                           title='Profiler results for {}'.format(details['endpoint']), body=body)



"""
    Contains all menu items that can be found in the Dashboard-menu on the left.
    The endpoints are the following:
    - Overview: shows a quick overview of every endpoint.
    - Hourly API Utilization: shows a heatmap of the number of requests.
    - Multi Version API Utilization: shows how much a certain version is used.
    - Daily API Utilization: shows how many requests are received per day.
    - API Performance: Shows performance box plots for every endpoint.
"""

from flask_monitoringdashboard.views.dashboard.endpoints import endpoints
from flask_monitoringdashboard.views.dashboard.heatmap import hourly_load
from flask_monitoringdashboard.views.dashboard.overview import overview
from flask_monitoringdashboard.views.dashboard.requests import requests
from flask_monitoringdashboard.views.dashboard.version_usage import version_usage




def log(string):
    """
    Only print output if this is specified in the configuration
    :param string: string to be printed
    """
    from flask_monitoringdashboard import config
    if config.enable_logging:
        print(string)



import datetime

from flask import render_template

from flask_monitoringdashboard import blueprint, config
from flask_monitoringdashboard.core.timezone import to_local_datetime
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.core.auth import admin_secure
from flask_monitoringdashboard.core.utils import get_details


@blueprint.route('/configuration', methods=['GET', 'POST'])
@admin_secure
def configuration():
    with session_scope() as db_session:
        details = get_details(db_session)
    details['first-request'] = to_local_datetime(datetime.datetime.fromtimestamp(details['first-request']))
    details['first-request-version'] = to_local_datetime(datetime.datetime.
                                                         fromtimestamp(details['first-request-version']))
    return render_template('fmd_config.html', details=details, config=config)



from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.colors import get_color
from flask_monitoringdashboard.core.forms import get_slider_form
from flask_monitoringdashboard.core.info_box import get_plot_info
from flask_monitoringdashboard.core.plot import get_layout, get_figure, boxplot
from flask_monitoringdashboard.database import session_scope, TestEndpoint
from flask_monitoringdashboard.database.count import count_test_builds, count_builds_endpoint
from flask_monitoringdashboard.database.count_group import get_value, count_times_tested, get_latest_test_version
from flask_monitoringdashboard.database.data_grouped import get_test_data_grouped
from flask_monitoringdashboard.database.tested_endpoints import get_tested_endpoint_names
from flask_monitoringdashboard.database.tests import get_test_suites, get_travis_builds, \
    get_endpoint_measurements_job, get_suite_measurements, get_last_tested_times, get_endpoint_measurements

AXES_INFO = '''The X-axis presents the execution time in ms. The Y-axis presents the
Travis builds of the Flask application.'''

CONTENT_INFO = '''In this graph, it is easy to compare the execution time of the different builds
to one another. This information may be useful to validate which endpoints need to be improved.'''


@blueprint.route('/test_build_performance', methods=['GET', 'POST'])
@secure
def test_build_performance():
    """
    Shows the performance results for the complete test runs of a number of Travis builds.
    :return:
    """
    with session_scope() as db_session:
        form = get_slider_form(count_test_builds(db_session), title='Select the number of builds')
    graph = get_boxplot_tests(form=form)
    return render_template('fmd_dashboard/graph.html', graph=graph, title='Per-Build Test Performance',
                           information=get_plot_info(AXES_INFO, CONTENT_INFO), form=form)


@blueprint.route('/endpoint_build_performance', methods=['GET', 'POST'])
@secure
def endpoint_build_performance():
    """
    Shows the performance results for the endpoint hits of a number of Travis builds.
    :return:
    """
    with session_scope() as db_session:
        form = get_slider_form(count_builds_endpoint(db_session), title='Select the number of builds')
    graph = get_boxplot_endpoints(form=form)
    return render_template('fmd_dashboard/graph.html', graph=graph, title='Per-Build Endpoint Performance',
                           information=get_plot_info(AXES_INFO, CONTENT_INFO), form=form)


@blueprint.route('/testmonitor/<end>', methods=['GET', 'POST'])
@secure
def endpoint_test_details(end):
    """
    Shows the performance results for one specific unit test.
    :param end: the name of the unit test for which the results should be shown
    :return:
    """
    with session_scope() as db_session:
        form = get_slider_form(count_builds_endpoint(db_session), title='Select the number of builds')
    graph = get_boxplot_endpoints(endpoint=end, form=form)
    return render_template('fmd_testmonitor/endpoint.html', graph=graph, title='Per-Version Performance for ' + end,
                           information=get_plot_info(AXES_INFO, CONTENT_INFO), endp=end, form=form)


@blueprint.route('/testmonitor')
@secure
def testmonitor():
    """
    Gives an overview of the unit test performance results and the endpoints that they hit.
    :return:
    """
    from numpy import median

    with session_scope() as db_session:
        latest_version = get_latest_test_version(db_session)
        tests_latest = count_times_tested(db_session,
                                          TestEndpoint.app_version == latest_version)
        tests = count_times_tested(db_session)
        median_latest = get_test_data_grouped(db_session, median,
                                              TestEndpoint.app_version == latest_version)
        median = get_test_data_grouped(db_session, median)
        tested_times = get_last_tested_times(db_session)

        result = []
        for endpoint in get_tested_endpoint_names(db_session):
            result.append({
                'name': endpoint,
                'color': get_color(endpoint),
                'tests-latest-version': get_value(tests_latest, endpoint),
                'tests-overall': get_value(tests, endpoint),
                'median-latest-version': get_value(median_latest, endpoint),
                'median-overall': get_value(median, endpoint),
                'last-tested': get_value(tested_times, endpoint, default=None)
            })

        return render_template('fmd_testmonitor/testmonitor.html', result=result)


def get_boxplot_tests(form=None):
    """
    Generates a box plot visualization for the unit test performance results.
    :param form: the form that can be used for showing a subset of the data
    :return:
    """
    trace = []
    with session_scope() as db_session:
        if form:
            suites = get_test_suites(db_session, limit=form.get_slider_value())
        else:
            suites = get_test_suites(db_session)

        if not suites:
            return None
        for s in suites:
            values = get_suite_measurements(db_session, suite=s)
            trace.append(boxplot(values=values, label='{} -'.format(s)))

        layout = get_layout(
            xaxis={'title': 'Execution time (ms)'},
            yaxis={'title': 'Travis Build', 'autorange': 'reversed'}
        )

        return get_figure(layout=layout, data=trace)


def get_boxplot_endpoints(endpoint=None, form=None):
    """
    Generates a box plot visualization for the unit test endpoint hits performance results.
    :param endpoint: if specified, generate box plot for a specific endpoint, otherwise, generate for all tests
    :param form: the form that can be used for showing a subset of the data
    :return:
    """
    trace = []
    with session_scope() as db_session:
        if form:
            ids = get_travis_builds(db_session, limit=form.get_slider_value())
        else:
            ids = get_travis_builds(db_session)

        if not ids:
            return None
        for id in ids:
            if endpoint:
                values = get_endpoint_measurements_job(db_session, name=endpoint, job_id=id)
            else:
                values = get_endpoint_measurements(db_session, suite=id)

            trace.append(boxplot(values=values, label='{} -'.format(id)))

        layout = get_layout(
            xaxis={'title': 'Execution time (ms)'},
            yaxis={'title': 'Travis Build', 'autorange': 'reversed'}
        )

        return get_figure(layout=layout, data=trace)



import unittest

from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, login, get_test_app


class TestLogin(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        self.app = get_test_app()

    def test_get_login(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        with self.app.test_client() as c:
            self.assertEqual(200, c.get('dashboard/login').status_code)
            login(c)
            self.assertEqual(302, c.get('dashboard/login').status_code)

    def test_incorrect_login(self):
        """
            Try whether logging with incorrect credentials returns the login page
        """
        args = {'name': 'admin', 'password': 'wrong'}
        with self.app.test_client() as c:
            self.assertIn('formLogin', c.post('dashboard/login', data=args).data.decode())

    def test_correct_login(self):
        """
            Try whether logging with correct credentials does not return the login page
        """
        args = {'name': 'admin', 'password': 'admin'}
        with self.app.test_client() as c:
            self.assertNotIn('formLogin', c.post('dashboard/login', data=args).data.decode())

    def test_logout(self):
        """
            Just retrieve the content and check if nothing breaks
        """
        with self.app.test_client() as c:
            self.assertEqual(302, c.get('dashboard/logout').status_code)



"""
    This file contains all unit tests for the endpoint-table in the database. (Corresponding to the file:
    'flask_monitoringdashboard/database/endpoint.py')
    See info_box.py for how to run the test-cases.
"""

import unittest

from flask_monitoringdashboard.core.timezone import to_utc_datetime
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, NAME, TIMES


class TestEndpoint(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    def test_get_endpoint(self):
        """
            Test wheter the function returns the right values.
        """
        from flask_monitoringdashboard.database.endpoint import get_endpoint_by_name
        from flask_monitoringdashboard import config
        with session_scope() as db_session:
            endpoint = get_endpoint_by_name(db_session, NAME)
        self.assertEqual(endpoint.name, NAME)
        self.assertEqual(endpoint.monitor_level, 1)
        self.assertEqual(endpoint.version_added, config.version)

    def test_update_endpoint(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.endpoint import get_endpoint_by_name, update_endpoint
        with session_scope() as db_session:
            current_value = get_endpoint_by_name(db_session, NAME).monitor_level
            new_value = 1 if current_value != 1 else 2
            update_endpoint(db_session, NAME, new_value)
            self.assertEqual(get_endpoint_by_name(db_session, NAME).monitor_level, new_value)

    def test_update_last_accessed(self):
        """
            Test whether the function returns the right values.
        """
        import datetime
        time = datetime.datetime.utcnow()
        from flask_monitoringdashboard.database.endpoint import update_last_accessed, get_last_requested
        from flask_monitoringdashboard.database.count_group import get_value
        with session_scope() as db_session:
            update_last_accessed(db_session, NAME)
            result = get_value(get_last_requested(db_session), NAME)
            result_utc = to_utc_datetime(result)
            self.assertTrue((result_utc - time).seconds < 1)

    def test_endpoints(self):
        """
            Test whether the function returns the right values.
        """
        from flask_monitoringdashboard.database.endpoint import get_endpoints
        from flask_monitoringdashboard import config
        with session_scope() as db_session:
            result = get_endpoints(db_session)
            self.assertEqual(len(result), 1)
            self.assertEqual(result[0].name, NAME)
            self.assertEqual(result[0].monitor_level, 1)
            self.assertEqual(result[0].version_added, config.version)
            self.assertEqual(result[0].last_requested, TIMES[0])



import unittest

from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, get_test_app, \
    test_admin_secure, NAME


class TestTestMonitor(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        self.app = get_test_app()

    def test_build_performance(self):
        test_admin_secure(self, 'test_build_performance')

    def test_endpoint_build_performance(self):
        test_admin_secure(self, 'endpoint_build_performance')

    def test_testmonitor(self):
        test_admin_secure(self, 'testmonitor')

    def test_endpoint_test_details(self):
        test_admin_secure(self, 'testmonitor/{}'.format(NAME))



import datetime

from sqlalchemy import func

from flask_monitoringdashboard.core.timezone import to_utc_datetime
from flask_monitoringdashboard.database import Request, TestEndpoint, Endpoint


def get_latest_test_version(db_session):
    """
    Retrieves the latest version of the user app that was tested.
    :param db_session: session for the database
    :return latest test version
    """
    latest_time = db_session.query(func.max(TestEndpoint.time_added)).one()[0]
    if latest_time:
        return db_session.query(TestEndpoint.app_version).filter(TestEndpoint.time_added == latest_time).one()[0]
    return None


def count_rows_group(db_session, column, *criterion):
    """
    Count the number of rows of a specified column
    :param db_session: session for the database
    :param column: column to count
    :param criterion: where-clause of the query
    :return list with the number of rows per endpoint
    """
    return db_session.query(Request.endpoint_id, func.count(column)). \
        filter(*criterion).group_by(Request.endpoint_id).all()


def get_value(list, name, default=0):
    """
    :param list: must be structured as: [(a, b), (c, d), ..]
    :param name: name to filter on, e.g.: if name == a, it returns b
    :param default: returned if the name was not found in the list
    :return value corresponding to the name in the list.
    """
    for key, value in list:
        if key == name:
            return value
    return default


def count_requests_group(db_session, *where):
    """
    Return the number of hits for all endpoints (possible with more filter arguments).
    :param db_session: session for the database
    :param where: additional arguments
    """
    return count_rows_group(db_session, Request.id, *where)


def count_times_tested(db_session, *where):
    """ Return the number of tests for an endpoint (possibly with more filter arguments).
    :param db_session: session for the database
    :param where: additional arguments
    """
    result = db_session.query(Endpoint.name, func.count(TestEndpoint.endpoint_id)).join(
        TestEndpoint.endpoint).filter(*where).group_by(Endpoint.name).all()
    return result


def count_requests_per_day(db_session, list_of_days):
    """ Return the number of hits for all endpoints per day.
    :param db_session: session for the database
    :param list_of_days: list with datetime.datetime objects. """
    result = []
    for day in list_of_days:
        dt_begin = to_utc_datetime(datetime.datetime.combine(day, datetime.time(0, 0, 0)))
        dt_end = dt_begin + datetime.timedelta(days=1)

        result.append(count_rows_group(db_session, Request.id, Request.time_requested >= dt_begin,
                                       Request.time_requested < dt_end))
    return result
